""" Monkey patches for loopy.symbolic

Use this module to insert pymbolic nodes and the likes.
"""
from dune.perftool.error import PerftoolError
from pymbolic.mapper.substitutor import make_subst_func

import loopy as lp
import pymbolic.primitives as prim
from pytools import ImmutableRecord
from six.moves import intern


#
# Pymbolic nodes to insert into the symbolic language understood by loopy
#


class SumfactKernel(ImmutableRecord, prim.Variable):
    def __init__(self,
                 a_matrices=None,
                 buffer=None,
                 stage=1,
                 preferred_position=None,
                 restriction=0,
                 within_inames=(),
                 input=None,
                 padding=frozenset(),
                 index=None,
                 insn_dep=frozenset(),
                 coeff_func=None,
                 element=None,
                 component=None,
                 accumvar=None,
                 ):
        """Create a sum factorization kernel

        Sum factorization can be written as

        Y = R_{d-1} (A_{d-1} * ... * R_0 (A_0 * X)...)

        with:
        - X: Input rank d tensor of dimension n_0 x ... x n_{d-1}
        - Y: Output rank d tensor of dimension m_0 x ... x m_{d-1}
        - A_l: Values of 1D basis evaluations at quadrature points in l
               direction, matrix of dimension m_l x n_l
        - R_l: Transformation operator that permutes the underlying data
               vector of the rank d tensor in such a way that the fastest
               direction gets the slowest direction

        In the l'th step we have the following setup:
        - A_l: Matrix of dimensions m_l x n_l
        - X_l: Rank d tensor of dimensions n_l x ... x n_{d-1} x m_0 x ... x m_{l-1}
        - R_l: Transformation operator

        Looking at the indizes the following will happen:
        X --> [n_l,...,n_{d-1},m_0,...,m_{l-1}]
        A_l * X --> [m_l,n_l] * [n_l, ...] = [m_l,n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]
        R_l (A_l*X) --> [n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]

        So the multiplication with A_l is a reduction over one index and
        the transformation brings the next reduction index in the fastest
        position.

        It can make sense to permute the order of directions. If you have
        a small m_l (e.g. stage 1 on faces) it is better to do direction l
        first. This can be done by:

        - Permuting the order of the A matrices.
        - Permuting the input tensor.
        - Permuting the output tensor (this assures that the directions of
          the output tensor are again ordered from 0 to d-1).

        Note, that you will typically *not* set all of the below arguments,
        but only some. The vectorization strategy may set others for you.
        The only argument really needed in all cases is a_matrices.

        Arguments:
        ----------
        a_matrices: A tuple of AMatrix instances
            The list of tensors to be applied to the input.
            Order of application is from 0 up.
        buffer: A string identifying the flip flop buffer in use
            for intermediate results. The memory is expected to be
            pre-initialized with the input or you have to provide
            direct_input (FastDGGridOperator).
        stage: 1 or 3
        insn_dep: An instruction ID that the first issued instruction
            should depend upon. All following ones will depend on each
            other.
        within_inames: Instructions will be executed within those
            inames (needed for stage 3 in jacobians).
        preferred_position: Will be used in the dry run to order kernels
            when doing vectorization e.g. (dx u,dy u,dz u, u).
        restriction: Restriction for faces values.
        input: The name to use for the input temporary
        padding: a set of slots in the input temporary to be padded
        index: The slot in the input temporary dedicated to this kernel
        coeff_func: The function to call to get the input coefficient
        element: The UFL element
        component: The treepath to the correct component of above element
        accumvar: The accumulation variable to accumulate into
    """
        # Check the input and apply defaults where necessary
        assert isinstance(a_matrices, tuple)
        from dune.perftool.sumfact.amatrix import AMatrixBase
        assert all(isinstance(m, AMatrixBase) for m in a_matrices)

        assert stage in (1, 3)

        if preferred_position is not None:
            assert isinstance(preferred_position, int)

        if stage == 3:
            assert isinstance(restriction, tuple)

        assert isinstance(within_inames, tuple)

        assert isinstance(insn_dep, frozenset)

        ImmutableRecord.__init__(self,
                                 a_matrices=a_matrices,
                                 buffer=buffer,
                                 stage=stage,
                                 preferred_position=preferred_position,
                                 restriction=restriction,
                                 within_inames=within_inames,
                                 input=input,
                                 padding=padding,
                                 index=index,
                                 insn_dep=insn_dep,
                                 coeff_func=coeff_func,
                                 element=element,
                                 component=component,
                                 accumvar=accumvar,
                                 )

        prim.Variable.__init__(self, "SUMFACT")

    #
    # The methods/fields needed to get a well-formed pymbolic node
    #
    def __getinitargs__(self):
        return (self.a_matrices, self.buffer, self.stage, self.preferred_position, self.restriction, self.within_inames, self.input, self.padding, self.index, self.insn_dep, self.coeff_func, self.element, self.component, self.accumvar)

    def stringifier(self):
        return lp.symbolic.StringifyMapper

    init_arg_names = ("a_matrices", "buffer", "stage", "preferred_position", "restriction", "within_inames", "input", "padding", "index", "insn_dep", "coeff_func", "element", "component", "accumvar")

    mapper_method = "map_sumfact_kernel"

    #
    # Some convenience methods to extract information about the sum factorization kernel
    #

    @property
    def length(self):
        return len(self.a_matrices)

    @property
    def vectorized(self):
        return next(iter(self.a_matrices)).vectorized

    @property
    def transposed(self):
        return next(iter(self.a_matrices)).transpose

    @property
    def cache_key(self):
        """ The cache key that can be used in generation magic,
        Any two sum factorization kernels having the same cache_key
        are realized simulatneously!
        """
        return hash((self.a_matrices, self.restriction, self.stage, self.buffer))

    @property
    def vec_index(self):
        """ A tuple with the vector index

        Defaults to the empty tuple to allow addition to any
        existing index tuple """
        if self.index is not None:
            return (self.index,)
        else:
            return ()

    @property
    def flat_input_shape(self):
        """ The 'flat' input tensor shape """
        from pytools import product
        shape = (product(mat.cols for mat in self.a_matrices),)
        if self.vectorized:
            shape = shape + (4,)
        return shape

    @property
    def quadrature_shape(self):
        """ The shape of a temporary for the quadrature points

        Takes into account the lower dimensionality of faces and vectorization.
        """
        if self.transposed:
            shape = tuple(mat.cols for mat in self.a_matrices if mat.face is None)
        else:
            shape = tuple(mat.rows for mat in self.a_matrices if mat.face is None)
        if self.vectorized:
            shape = shape + (4,)
        return shape

    @property
    def quadrature_dimtags(self):
        """ The dim_tags of a temporary for the quadrature points

        Takes into account the lower dimensionality of faces and vectorization.
        """
        tags = ["f"] * len(self.quadrature_shape)
        if self.vectorized:
            tags[-1] = 'c'
        return ",".join(tags)

    @property
    def dof_shape(self):
        """ The shape of a temporary for the degrees of freedom

        Takes into account vectorization.
        """
        shape = tuple(mat.rows for mat in self.a_matrices)
        if self.vectorized:
            shape = shape + (4,)
        return shape

    @property
    def dof_dimtags(self):
        """ The dim_tags of a temporary for the degrees of freedom

        Takes into account vectorization.
        """
        tags = ["f"] * len(self.dof_shape)
        if self.vectorized:
            tags[-1] = 'vec'
        return ",".join(tags)

    @property
    def output_shape(self):
        if self.stage == 1:
            return self.quadrature_shape
        else:
            return self.dof_shape

    @property
    def output_dimtags(self):
        if self.stage == 1:
            return self.quadrature_dimtags
        else:
            return self.dof_dimtags


class FusedMultiplyAdd(prim.Expression):
    """ Represents an FMA operation """

    init_arg_names = ("mul_op1", "mul_op2", "add_op")

    def __init__(self, mul_op1, mul_op2, add_op):
        self.mul_op1 = mul_op1
        self.mul_op2 = mul_op2
        self.add_op = add_op

    def __getinitargs__(self):
        return (self.mul_op1, self.mul_op2, self.add_op)

    def stringifier(self):
        return lp.symbolic.StringifyMapper

    mapper_method = intern("map_fused_multiply_add")


#
# Mapper methods to monkey patch into the visitor base classes!
#


def identity_map_sumfact_kernel(self, expr, *args):
    return expr


def walk_map_sumfact_kernel(self, expr, *args):
    self.visit(expr)


def stringify_map_sumfact_kernel(self, expr, *args):
    return "SUMFACT"


def dependency_map_sumfact_kernel(self, expr):
    return set()


def needs_resolution(self, expr):
    raise PerftoolError("SumfactKernel node is a placeholder and needs to be removed!")


def identity_map_fused_multiply_add(self, expr, *args):
    return FusedMultiplyAdd(self.rec(expr.mul_op1, *args),
                            self.rec(expr.mul_op2, *args),
                            self.rec(expr.add_op, *args),
                            )


def walk_map_fused_multiply_add(self, expr, *args):
    if not self.visit(expr):
        return

    self.rec(expr.mul_op1, *args)
    self.rec(expr.mul_op2, *args)
    self.rec(expr.add_op, *args)


def stringify_map_fused_multiply_add(self, expr, enclosing_prec):
    from pymbolic.mapper.stringifier import PREC_NONE
    return "fma(%s*%s+%s)" % (self.rec(expr.mul_op1, PREC_NONE),
                              self.rec(expr.mul_op2, PREC_NONE),
                              self.rec(expr.add_op, PREC_NONE))


def dependency_map_fused_multiply_add(self, expr):
    return self.combine((self.rec(expr.mul_op1),
                         self.rec(expr.mul_op2),
                         self.rec(expr.add_op)
                         ))


def type_inference_fused_multiply_add(self, expr):
    return self.rec(expr.mul_op1)


def vectorizability_map_fused_multiply_add(self, expr):
    return all((self.rec(expr.mul_op1), self.rec(expr.mul_op2), self.rec(expr.add_op)))


#
# Do the actual monkey patching!!!
#

# SumfactKernel node
lp.symbolic.IdentityMapper.map_sumfact_kernel = identity_map_sumfact_kernel
lp.symbolic.SubstitutionMapper.map_sumfact_kernel = lp.symbolic.SubstitutionMapper.map_variable
lp.symbolic.WalkMapper.map_sumfact_kernel = walk_map_sumfact_kernel
lp.symbolic.StringifyMapper.map_sumfact_kernel = stringify_map_sumfact_kernel
lp.symbolic.DependencyMapper.map_sumfact_kernel = dependency_map_sumfact_kernel
lp.target.c.codegen.expression.ExpressionToCExpressionMapper.map_sumfact_kernel = needs_resolution
lp.type_inference.TypeInferenceMapper.map_sumfact_kernel = needs_resolution

# FusedMultiplyAdd node
lp.symbolic.IdentityMapper.map_fused_multiply_add = identity_map_fused_multiply_add
lp.symbolic.WalkMapper.map_fused_multiply_add = walk_map_fused_multiply_add
lp.symbolic.StringifyMapper.map_fused_multiply_add = stringify_map_fused_multiply_add
lp.symbolic.DependencyMapper.map_fused_multiply_add = dependency_map_fused_multiply_add
lp.type_inference.TypeInferenceMapper.map_fused_multiply_add = type_inference_fused_multiply_add
lp.expression.VectorizabilityChecker.map_fused_multiply_add = vectorizability_map_fused_multiply_add

#
# Some helper functions!
#


def substitute(expr, replacemap):
    """ A replacement for pymbolic.mapper.subsitutor.substitute which is aware of all
    monkey patches etc.
    """
    return lp.symbolic.SubstitutionMapper(make_subst_func(replacemap))(expr)
